/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"
#include "dvbcompat.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"
#include "io.h"

/*
 * Input service details from szap2 channels.conf
 */
string
io::inszap(class Satellite *sat, class Transponder *tp, class Service *service,
	string zapfile, string ident)
{
	string match = "";

	if (access(zapfile.c_str(), R_OK) != 0)
	{
		string error = "Can't access zapfile [" + zapfile + "]";
		throw error;
	}

	ifstream ifs(zapfile.c_str(), ios::in | ios::binary);
	if (!ifs.is_open())
	{
		string error = "Can't open zapfile [" + zapfile + "]";
		throw error;
	}

	char cbuf[BUFLEN];
	regex_t r;

	util::regexPrepare(&r, ident.c_str());

	do
	{
		ifs.getline(cbuf, BUFLEN);

		if(ifs.eof())
			break;

		vector<string> parts = util::split((string)cbuf, ':', 0);

		if (parts.size() != 8)
			continue;

		match = parts[0];
		vector<string> matches = util::regex(&r, parts[0]);

		if (matches.size())
		{
			dprintf(30, "IDENT = %s\n", match.c_str());

			/* LNB */
			sat->lnb.dsec_port = util::toInt(parts[3]);
			dprintf(30, "DSEC PORT = %d\n", sat->lnb.dsec_port);

			/* Transponder */
			tp->frequency = util::toInt(parts[1]) * 1000;
			dprintf(30, "FREQUENCY = %d\n", tp->frequency);

			tp->symbol_rate = util::toInt(parts[4]) * 1000;
			dprintf(30, "SYMBOL RATE = %d\n", tp->symbol_rate);

			const char *p = parts[2].c_str();
			tp->voltage = dvbcompat::polarisation2voltaget(*p++);
			dprintf(30, "POLARISATION = %s\n", dvbcompat::voltaget2cstr(tp->voltage));

			while (*p)
			{
				char c = *p++;

				string parm = "";
				while (*p >= '0' && *p <= '9')
					parm += *p++;

				switch (c)
				{
					case 'C': tp->coderate = dvbcompat::szap2coderatet(parm);
						  dprintf(30, "CODE RATE = %s\n", dvbcompat::coderatet2cstr(tp->coderate));
						  break;
					case 'M': tp->modulation = dvbcompat::szap2modulationt(parm);
						  dprintf(30, "MODULATION = %s\n", dvbcompat::modulationt2cstr(tp->modulation));
						  break;
					case 'O': tp->rolloff = dvbcompat::szap2rollofft(parm);
						  dprintf(30, "ROLL OFF = %s\n", dvbcompat::rollofft2cstr(tp->rolloff));
						  break;
					case 'S': tp->delsys = dvbcompat::szap2delsyst(parm);
						  dprintf(30, "DELIVERY SYSTEM = %s\n", dvbcompat::delsyst2cstr(tp->delsys));
						  break;
				}
			}

			/* Service */
			service->video = util::toInt(parts[5]);
			dprintf(30, "VPID = %d\n", service->video);

			service->audio[0] = util::toInt(parts[6]);
			dprintf(30, "APID = %d\n", service->audio[0]);

			service->sid = util::toInt(parts[7]);
			dprintf(30, "SID = %d\n", service->sid);

			break;
		}
		else
			match = "";
	}
	while( true );

	ifs.close();

	return match;
}

/*
 * Output service details for szap2 channels.conf
 */
void
io::outszap(class Satellite *sat, class Transponder *tp, class Service *service)
{
	const char charfilter[] = ";:\n";
	string newprovider, newservice;
	
	newprovider = cstring::stringcpyf(service->provider, charfilter);
	newservice = cstring::stringcpyf(service->service, charfilter);

	if (newprovider.length())
	{
		if (newservice.length())
			/* sid=service;provider */
			cout << service->sid << "=" << newservice << ";" << newprovider;
		else
			/* sid;provider */
			cout << service->sid << ";" << newprovider;
	}
	else
	{
		if (newservice.length())
			/* sid=service */
			cout << service->sid << "=" << newservice;
		else
			/* sid */
			cout << service->sid;
	}

	cout << ":" << tp->frequency / 1000;
	cout << ":" << dvbcompat::voltaget2cstr(tp->voltage);
	cout << "C" << dvbcompat::coderatet2szap(tp->coderate);
	cout << "M" << dvbcompat::modulationt2szap(tp->modulation);
	cout << "O" << dvbcompat::rollofft2szap(tp->rolloff);
	cout << "S" << dvbcompat::delsyst2szap(tp->delsys);
	cout << ":" << sat->lnb.dsec_port / 1;
	cout << ":" << tp->symbol_rate / 1000;
	cout << ":" << service->video;
	cout << ":" << service->audio[0];
	cout << ":" << service->sid << endl;
}

/*
 * Output service details for vdr channels.conf
 */
void
io::outvdr(class Satellite *sat, class Transponder *tp, class Service *service)
{
	const char charfilter[] = ";:\n";
	string newprovider, newservice;
	
	newprovider = cstring::stringcpyf(service->provider, charfilter);
	newservice = cstring::stringcpyf(service->service, charfilter);

	if (newprovider.length())
	{
		if (newservice.length())
			/* service;provider */
			cout << newservice << ";" << newprovider;
		else
			/* [sid];provider */
			cout << "[" << service->sid << "];" << newprovider;
	}
	else
	{
		if (newservice.length())
			/* service */
			cout << newservice;
		else
			/* [sid] */
			cout << "[" << service->sid << "]";
	}

	cout << ":" << tp->frequency / 1000;
	cout << ":" << dvbcompat::voltaget2cstr(tp->voltage);
	cout << "C" << dvbcompat::coderatet2szap(tp->coderate);
	cout << "M" << dvbcompat::modulationt2szap(tp->modulation);
	cout << "O" << dvbcompat::rollofft2szap(tp->rolloff);
	cout << "S" << dvbcompat::delsyst2szap(tp->delsys);
	cout <<":S" << util::angleToString(sat->lnb.position);
	cout << ":" << tp->symbol_rate / 1000;
	cout << ":" << service->video;
	cout << ":" << service->serialiseAudio();
	cout << ":" << service->text;
	cout << ":" << service->serialiseCaid();
	cout << ":" << service->sid;
	cout << ":" << tp->onid;
	cout << ":" << tp->tid;
	cout <<":0" << endl; /* RID */
}

/*
 * Output pre-scan satellite and transponder details
 */
void
io::outsat(class Satellite *sat)
{
	cout << sat->name << endl;
	for (list<Transponder>::iterator tp = sat->tp_list.begin(); tp != sat->tp_list.end(); ++tp)
	{
		cout    << "    " << tp->frequency/1000 << " "
			<< dvbcompat::polarisation2cstr(tp->polarisation) << " "
			<< dvbcompat::delsyst2cstr(tp->delsys) <<  " "
			<< dvbcompat::modulationt2cstr(tp->modulation) << " "
			<< tp->symbol_rate/1000 << " "
			<< dvbcompat::coderatet2cstr(tp->coderate) << endl;
	}
	cout << endl;
}

/*
 * Output transponder details for inspection (in `KingOfSat' format)
 */
void
io::outkos(class Satellite *sat, class Transponder *tp, string error)
{
	cout	<< sat->name << " "
		<< (double)tp->frequency/1000.0 << " "
		<< dvbcompat::polarisation2cstr(tp->polarisation, true) << " "
		<< dvbcompat::delsyst2cstr(tp->delsys) << " "
		<< dvbcompat::modulationt2cstr(tp->modulation) << " "
		<< tp->symbol_rate/1000 << " "
		<< dvbcompat::coderatet2cstr(tp->coderate) << " "
		<< "ONID:" << tp->onid << " "
		<< "TID:" << tp->tid << " "
		<< "AGC:" << PERCENT(sat->lnb.signal) << "% "
		<< "SNR:" << PERCENT(sat->lnb.snr) << "% " << endl;

	if (error != "")
	{
		cout << "    " << error << endl << endl;
	}
	else /* XXX check for no services */
	{
		cout << "    " << resetiosflags(ios::right) << setiosflags(ios::left) <<
			setw(33) << "SERVICE" << setw(25) << "PROVIDER" << "CA SID    VPID  APID  PMT   PCR    TXT\n";

		for (vector<class Service>::iterator si = tp->service.begin(); si < tp->service.end(); si++)
		{
			cout << si->typestr << "  "
				<< cstring::stringcpyp(si->service, 32) << " "
				<< cstring::stringcpyp(si->provider, 25) << " "
				<< ((si->caid.size()) ? "X" : " ") << " ";

			cout << resetiosflags(ios::right) << setiosflags(ios::left)  << setw(7) << si->sid;
			cout << setw(6) << si->video;
			cout << setw(6) << si->audio[0];
			cout << setw(6) << si->pmt_pid;
			cout << setw(6) << si->pcr_pid;
			cout << resetiosflags(ios::left)  << setiosflags(ios::right) << setw(4) << si->text << endl;

			for (unsigned int audix = 1; audix < si->audio.size(); audix++)
			{
				cout << resetiosflags(ios::right) << setiosflags(ios::left);
				cout << setw(78) << " " << si->audio[audix] << endl;
			}
		}
		cout << endl;
	}
}
