/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"
#include "io.h"

Lnb::Lnb(void)
{
	fefd = -1;
	defd = -1;
	defdv = -1;
	defda = -1;
	low = 9750000;
	high = 10600000;
	threshold = 11700000;
	polarisation = 'L';	/* LINEAR / CIRCULAR */
	position = 0;
	frequency = 0;		/* Tuning frequency */
	dsec_port = 0;
	dsec_repeat = 2;
	dsec_burst = 1;
	tuning_delay = 5;
	phy_adapter = 0;
	tuned = false;
	signal = 0;
	snr = 0;
	ber = 0;
	unc = 0;
	phy_adapter = 0;
}

Lnb::~Lnb(void)
{
	close();
}

void
Lnb::open(int mode)
{
	struct dvb_frontend_info fe_info;
	int err;
	stringstream fname;

	close();

	fname << "/dev/dvb/adapter" << phy_adapter << "/frontend0";
	fefd = ::open(fname.str().c_str(), O_RDWR | O_NONBLOCK);
	if (fefd < 0)
	{
		string error = "Can't open DVB adapter frontend [" + fname.str() +"]";
		throw error;
	}

	err = ioctl(fefd, FE_GET_INFO, &fe_info);
	if (err)
	{
		string error = "Can't get DVB adapter frontend info [" + fname.str() + "]";
		throw error;
	}

	if (fe_info.type != FE_QPSK)
	{
		string error = "DVB adapter frontend is not DVB-S/S2 [" + fname.str() + "]";
		throw error;
	}

	err = ioctl(fefd, FE_SET_FRONTEND_TUNE_MODE, mode);
	if (err)
	{
		string error = "Can't set DVB adapter frontend mode [" + fname.str() + "]";
		throw error;
	}

	fname.str("");
	fname << "/dev/dvb/adapter" << phy_adapter << "/demux0";
	defd = ::open(fname.str().c_str(), O_RDWR | O_NONBLOCK);
	if (defd < 0)
	{
		string error = "Can't open DVB adapter demux [" + fname.str() + "]";
		throw error;
	}

	defdv = ::open(fname.str().c_str(), O_RDWR);
	if (defdv < 0)
	{
		string error = "Can't open DVB adapter video demux [" + fname.str() + "]";
		throw error;
	}
	
	defda = ::open(fname.str().c_str(), O_RDWR);
	if (defda < 0)
	{
		string error = "Can't open DVB adapter audio demux [" + fname.str() + "]";
		throw error;
	}
}

void
Lnb::close(void)
{
	if (defda != -1)
	{
		::close(defda);
		defda = -1;
	}
	if (defdv != -1)
	{
		::close(defdv);
		defdv = -1;
	}
	if (defd != -1)
	{
		::close(defd);
		defd = -1;
	}
	if (fefd != -1)
	{
		::close(fefd);
		fefd = -1;
	}
}

int
Lnb::fcntl(int flag, bool set)
{
	int err, flags;

	err = ::fcntl(fefd, F_GETFL);
	if (err < 0)
		return err;

	if (set)
		flags = err | flag;
	else
		flags = err & ~flag;

	err = ::fcntl(fefd, F_SETFL, flags);
	if (err < 0)
		return err;

	return 0;
}

int
Lnb::read(unsigned char *cbuf)
{
	struct timeval timeval;
	fd_set fdset;
	int err;
	
	bzero(cbuf, io::BUFLEN);
	while (1)
	{
		timeval.tv_sec = TIMEOUT;
		timeval.tv_usec = 0;
		FD_ZERO(&fdset);
		FD_SET(defd, &fdset);

		err = select(defd + 1, &fdset, NULL, NULL, &timeval);
		if (err < 0)
		{
			if (errno == EINTR)
				continue;
			return -errno;
		}

		if (err == 0)
			return -ETIMEDOUT;

		if (FD_ISSET(defd, &fdset) == 0)
			return -ETIMEDOUT;

		err = ::read(defd, cbuf, io::BUFLEN);
		if (err < 0)
		{
			if (errno == EINTR)
				continue;
			return -errno;
		}
		return err;
	}
}

/*
 * Start SCT
 */
void
Lnb::start(int pid, int tableid)
{
	struct dmx_sct_filter_params sct;
	int err;

	bzero(&sct, sizeof(struct dmx_sct_filter_params));
	sct.flags = DMX_IMMEDIATE_START | DMX_CHECK_CRC;
	sct.pid = pid;
	sct.filter.filter[0] = tableid;
	sct.filter.mask[0] = 0xff;
	sct.timeout = 0;
	err = ioctl(defd, DMX_SET_FILTER, &sct);
	if (err)
		throw "Can't set DVB adapter demux filter";
}

/*
 * Start PES
 */
void
Lnb::start(class Service *service)
{
	struct dmx_pes_filter_params pes;
	int err;

	dprintf(30, "VIDEO = %d\n", service->video);
	if (service->video)
	{
		err = ioctl(defdv, DMX_SET_BUFFER_SIZE, 65536);
		if (err)
			throw "Can't set DVB adapter demux buffer size";
		bzero(&pes, sizeof(struct dmx_pes_filter_params));
		pes.pid = service->video;
		pes.input = DMX_IN_FRONTEND;
		pes.output = DMX_OUT_TS_TAP;
		pes.pes_type = DMX_PES_VIDEO;
		pes.flags = DMX_IMMEDIATE_START;
		err = ioctl(defdv, DMX_SET_PES_FILTER, &pes);
		if (err)
			throw "Can't set DVB adapter demux filter";
	}
	dprintf(30, "AUDIO = %d\n", service->audio[0]);
	if (service->audio[0])
	{
		err = ioctl(defda, DMX_SET_BUFFER_SIZE, 65536);
		if (err)
			throw "Can't set DVB adapter demux buffer size";
		bzero(&pes, sizeof(struct dmx_pes_filter_params));
		pes.pid = service->audio[0];
		pes.input = DMX_IN_FRONTEND;
		pes.output = DMX_OUT_TS_TAP;
		pes.pes_type = DMX_PES_AUDIO;
		pes.flags = DMX_IMMEDIATE_START;
		err = ioctl(defda, DMX_SET_PES_FILTER, &pes);
		if (err)
			throw "Can't set DVB adapter demux filter";
	}
}

void
Lnb::stop(void)
{
	ioctl(defd, DMX_STOP);
}

/*
 * XXX remove or move this
 */
void
Lnb::spin(void)
{
	static int c_count = 0;
	static const char arrow[] = "|/-\\";
	int len = strlen(arrow);

	if (++c_count % (len - 1) == 0)
	{
		cout << arrow[ len - ( c_count % len ) - 1 ] << '\r';
		cout.flush();
	}
}

/*
 * Tune to transponder
 */
void
Lnb::tune(Transponder *tp)
{
	struct dvb_diseqc_master_cmd dsec;
	struct dvb_frontend_event ev;
	struct dtv_property args[10];
	struct dtv_properties seq;
	unsigned int band = low;
	unsigned int repeat = dsec_repeat;
	int err, retry;

	tuned = false;
	signal = 0;
	snr = 0;
	ber = 0;
	unc = 0;

	if (dsec_port > 3)
		throw "Invalid DiSEqC port";

	if (threshold && high && (tp->frequency >= threshold))
		band = high;

	dsec.msg[0] = 0xe0;
	dsec.msg[1] = 0x10;
	dsec.msg[2] = 0x38;
	dsec.msg[3] = 0xf0 | (dsec_port << 2) | (tp->voltage << 1) | (band == high);
	dsec.msg_len = 4;

	err = ioctl(fefd, FE_SET_TONE, SEC_TONE_OFF);
	if (err)
		throw "Can't set DVB adapter frontend tone";
	usleep(500);

	err = ioctl(fefd, FE_SET_VOLTAGE, tp->voltage);
	if (err)
		throw "Can't set DVB adapter frontend voltage";
	usleep(15000);
		    
	while (repeat--)
	{
		err = ioctl(fefd, FE_DISEQC_SEND_MASTER_CMD, &dsec);
		usleep(15000);
		if (dsec_burst)
		{
			err = ioctl(fefd, FE_DISEQC_SEND_BURST,
				(dsec_port & 1) ? SEC_MINI_B : SEC_MINI_A);
			if (err)
				throw "Can't set DVB adapter frontend burst";
			usleep(15000);
		}
		else
			usleep(28000);
#if 0
		if (repeat)
			usleep(100000);
#endif
	}

	if (band == high)
	{
		err = ioctl(fefd, FE_SET_TONE, SEC_TONE_ON);
		if (err)
			throw "Can't set DVB adapter frontend tone";
		usleep(500);
	}
	
	args[0].cmd = DTV_CLEAR;
	args[0].u.data = 0;

	args[1].cmd = DTV_DELIVERY_SYSTEM;
	args[1].u.data = tp->delsys;

	args[2].cmd = DTV_FREQUENCY;
	args[2].u.data = (tp->frequency < band)
		? (band - tp->frequency) : (tp->frequency - band);

	args[3].cmd = DTV_MODULATION;
	args[3].u.data = tp->modulation;

	args[4].cmd = DTV_SYMBOL_RATE;
	args[4].u.data = tp->symbol_rate;

	args[5].cmd = DTV_INNER_FEC;
	args[5].u.data = tp->coderate;

	args[6].cmd = DTV_ROLLOFF;
	args[6].u.data = tp->rolloff;

	args[7].cmd = DTV_PILOT;
	args[7].u.data = tp->pilot;

	args[8].cmd = DTV_INVERSION;
	args[8].u.data = INVERSION_AUTO;

	args[9].cmd = DTV_TUNE;
	args[9].u.data = 0;

	seq.props = args;
	seq.num = 10;

	do {
		err = ioctl(fefd, FE_GET_EVENT, &ev);
	}
	while (err != -1);

	err = ioctl(fefd, FE_SET_PROPERTY, &seq);
	if (err)
		throw "Can't set DVB adapter frontend properties (" + (string)strerror(errno) + ")";
	do {
		err = ioctl(fefd, FE_GET_EVENT, &ev);
	}
	while (ev.status != 0);

	retry = DELAY1S * tuning_delay / DELAY100MS;
	do
	{
		usleep(DELAY100MS);
		err = ioctl(fefd, FE_GET_EVENT, &ev);
	}
	while (!(ev.status & FE_HAS_LOCK) && retry--);

	if (ev.status & FE_HAS_LOCK)
	{
		tuned = true;
		retry = DELAY1S / DELAY100MS;
		while ((signal == 0 || snr == 0) && retry --)
		{
			usleep(DELAY100MS);
			err = ioctl(fefd, FE_READ_SIGNAL_STRENGTH, &signal);
			err = ioctl(fefd, FE_READ_SNR, &snr);
		}
		err = ioctl(fefd, FE_READ_BER, &ber);
		err = ioctl(fefd, FE_READ_UNCORRECTED_BLOCKS, &unc);
	}
}

/*
 * Tune to zap file ident
 */
void
Lnb::tune(class Satellite *sat, string zapfile, string ident, string command)
{
	class Transponder tp;
	class Service service;
	int err;

	string channel = io::inszap(sat, &tp, &service, zapfile, ident);
	if (channel == "")
		throw "Channel ident not found";
	
	tune(&tp);
	if (!tuned)
		throw "Can't tune to channel";

	start(&service);
	cout << "Tuned to '" << channel << "'" << endl;
	cout.flush();
		
	fcntl(O_NONBLOCK, false);
	if (command != "") {
		stringstream cmd;
		cmd << command << " /dev/dvb/adapter" << phy_adapter << "/dvr0";
		cout << "Exec '" << cmd.str() << "'\n";
		cout.flush();
		err = system(cmd.str().c_str());
		if (err) {
			/* Do nothing */
		}
	} else {
		struct dvb_frontend_event ev;
		cout << "Press CTRL-C to exit.\n";
		cout.flush();
		while (1) {
			err = ioctl(fefd, FE_GET_EVENT, &ev);
			if (err) {
				/* Do nothing */
			}
		}
	}
}
