/*
 * Copyright (c) 2004-2008 Darron M Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"
#include "matrix.h"
#include "sql.h"
extern "C" {
#include <mysql/mysql.h>
};
#include "mysql5.h"

/*
 * mysql 5.0.19+ interface class
 */

const char
Mysql5::SCHEMA[] = "/usr/local/etc/dvb2010.mysql";

Mysql5::Mysql5(const char *dbhost, const char *user, const char *password,
	const char *database_name, bool init, const char *schema_file)
{
	if (mysql_init(&db) == NULL)
		throw "error: cannot initialise mysql object";

	my_bool optval = 0;
	mysql_options(&db, MYSQL_OPT_RECONNECT, (const char *)&optval);

	this->dbhost = dbhost;
	this->user = user;
	this->password = password;
	this->database_name = database_name;
	this->schema_file = schema_file;

	if (init)
		initialise();
	else
		connect();
}

Mysql5::~Mysql5(void)
{
	mysql_close(&db);
}

/*
 * connect and intialise database
 */
void
Mysql5::initialise(void)
{
	dprintf(30, "CONNECTING TO DATABASE SERVER...\n");
	MYSQL *rc = mysql_real_connect(&db, dbhost.c_str(), user.c_str(),
		password.c_str(), NULL, 3306, NULL, 0UL);
	if (rc == NULL)
		throw mysql_error(&db); 
	dprintf(30, "CONNECTED\n");

	string query = (string) "DROP DATABASE IF EXISTS " + database_name;
	SQLcommand(query.c_str());
	query = (string) "CREATE DATABASE IF NOT EXISTS " + database_name +
		" CHARACTER SET utf8 COLLATE utf8_general_ci";
	SQLcommand(query.c_str());
	
	query = (string) "USE " + database_name;
	SQLcommand(query.c_str());

	SQLcommand("SET NAMES utf8");
	SQLcommand("SET CHARACTER SET utf8");
	
	file(schema_file.c_str());
	
	dprintf(30, "DATABASE READY\n");
}

/*
 * connect to database
 */
void
Mysql5::connect(void)
{
	dprintf(30, "CONNECTING TO DATABASE SERVER...\n");
	MYSQL *rc = mysql_real_connect(&db, dbhost.c_str(), user.c_str(),
		password.c_str(), database_name.c_str(), 3306, NULL, 0UL);
	if (rc == NULL)
		throw mysql_error(&db); 
	dprintf(30, "CONNECTED\n");

	SQLcommand("SET NAMES utf8");
	SQLcommand("SET CHARACTER SET utf8");

	dprintf(30, "DATABASE READY\n");
}

/*
 * reconnect to database
 */
void
Mysql5::reconnect(void)
{
	if (mysql_ping(&db) != 0)
		connect();
}

/*
 * peform query and return count
 *
 * nb. query must start as SELECT COUNT(*) FROM ...
 */
int
Mysql5::SQLcount(const char *query)
{
	dprintf(30, "QUERY [%s]\n", query);

	if (mysql_query(&db, query) != 0)
		throw mysql_error(&db);

	MYSQL_RES *mres = mysql_store_result(&db);
	if (mres == NULL)
		throw mysql_error(&db);

	int count=0;
	int nrow, ncol;
	MYSQL_ROW mrow;

	nrow= mysql_num_rows(mres);
	if (nrow != 1)
		goto nocount;
	
	mrow= mysql_fetch_row(mres);

	ncol= mysql_num_fields(mres);
	if (ncol != 1)
		goto nocount;

	count= atoi(mrow[0]);
nocount:mysql_free_result(mres);

	return count;
}
int
Mysql5::SQLcount(string query)
{
	return SQLcount(query.c_str());
}
int
Mysql5::SQLcount(stringstream &query)
{
	return SQLcount(query.str().c_str());
}

/*
 * perform a query that has no results
 */
void
Mysql5::SQLcommand(const char *query)
{
	dprintf(30, "QUERY [%s]\n", query);

	if (mysql_query(&db, query) != 0)
		throw mysql_error(&db);
}
void
Mysql5::SQLcommand(string query)
{
	SQLcommand(query.c_str());
}
void
Mysql5::SQLcommand(stringstream &query)
{
	SQLcommand(query.str().c_str());
}

/*
 * peform query and return number of rows and populate a result matrix
 */
int
Mysql5::SQLquery(const char *query, SqlMatrix *matrix)
{
	dprintf(30, "QUERY [%s]\n", query);

	if (mysql_query(&db, query) != 0)
		throw mysql_error(&db);

	MYSQL_RES *mres = mysql_store_result(&db);
	if (mres == NULL)
		throw mysql_error(&db);

	int nrow = mysql_num_rows(mres);
	if (nrow < 0)
	{
		mysql_free_result(mres);
		return 0;
	}
	int nfield = mysql_num_fields(mres);
	if (nfield < 0)
	{
		mysql_free_result(mres);
		return 0;
	}
	MYSQL_FIELD *field = mysql_fetch_fields(mres);
	for (int i = 0; i < nfield; i++)
		matrix->addFieldName(field[i].name);
	MYSQL_ROW mrow;
	int ncol = 0;
	for (int y = 0; y < nrow; y++)
	{
		mrow = mysql_fetch_row(mres);
		if (y == 0)
		{
			ncol = mysql_num_fields(mres);
			matrix->resize(ncol, nrow);
		}
		for (int x = 0; x < ncol; x++)
		{
			if (mrow[x] == NULL)
				matrix->set(x, y, (string)"(null)");
			else
				matrix->set(x, y, mrow[x]);
		}
	}
	mysql_free_result(mres);

	return nrow;
}
int
Mysql5::SQLquery(string query, SqlMatrix *matrix)
{
	return SQLquery(query.c_str(), matrix);
}
int
Mysql5::SQLquery(stringstream &query, SqlMatrix *matrix)
{
	return SQLquery(query.str().c_str(), matrix);
}

/*
 * escape strings to mysql requirements
 */
string
Mysql5::SQLescape(const char *cstr)
{
	char *temp = new char[2 * strlen(cstr) + 1];

	int newlen = mysql_real_escape_string(&db, temp, cstr, strlen(cstr));

	temp[newlen] = '\0';

	string newquery = temp;
	delete[] temp;

	return newquery;
}
string
Mysql5::SQLescape(string query)
{
	return SQLescape(query.c_str());
}

string
Mysql5::SQLnull(string field)
{
	return (string)"ISNULL(" + field + ")";
}

string
Mysql5::SQLquotefield(string field)
{
	return "`" + field + "`";
}
