/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "dvbcompat.h"
#include "cstring.h"
#include "util.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"
#include "matrix.h"
#include "sql.h"
#include "io.h"

const char
Satellite::SATXML[] = "/usr/local/etc/satellites.xml";

Satellite::Satellite(const char *xml)
{
	this->xml = xml;
	name[0] = '\0';
	depth = 0;
	tp_posfound = false;
	tp_num = 0;
}

namespace
{
	void
	start(void *data, const XML_Char *el __attribute__((unused)), const XML_Char **attr)
	{
		((Satellite *)data)->start(attr);
	}
}

void
Satellite::start(const XML_Char **attr)
{
	if (depth == 1)
	{
		string str;

		for (int i = 0; attr[i]; i += 2)
		{
			tp_posfound = false;
			if (strcasecmp(attr[i], "name") == 0)
				str = (string)attr[i+1];
			else if (strcasecmp(attr[i], "position") == 0)
			{
				if (lnb.position == util::toInt(attr[i+1]))
				{
					dprintf(20, "FOUND SATELLITE %s\n", str.c_str());
					name = str;
					tp_posfound = true;
				}
			}
		}
	}
	if (depth == 2 && tp_posfound)
	{
		Transponder tp;

		for (int i = 0; attr[i]; i += 2)
		{
			unsigned long tmp = util::toInt(attr[i+1]);

			if (strcasecmp(attr[i], "frequency") == 0)
				tp.frequency = tmp;
			else if (strcasecmp(attr[i], "symbol_rate") == 0)
				tp.symbol_rate = tmp;
			else if (strcasecmp(attr[i], "polarization") == 0)
			{
				tp.polarisation = tmp;
				tp.voltage = dvbcompat::polarisation2voltaget(tmp);
			}
			else if (strcasecmp(attr[i], "fec_inner") == 0)
				tp.coderate = dvbcompat::fec_inner2coderatet(tmp);
			else if (strcasecmp(attr[i], "system") == 0)
				tp.delsys = dvbcompat::system2delsyst(tmp);
			else if (strcasecmp(attr[i], "modulation") == 0)
				tp.modulation = dvbcompat::modulation2modulationt(tmp);
			else if (strcasecmp(attr[i], "roll_off") == 0)
				tp.rolloff = dvbcompat::rolloff2rollofft(tmp);
			else if (strcasecmp(attr[i], "pilot") == 0)
				tp.pilot = dvbcompat::pilot2pilott(tmp);
		}

		/* Default params for DVB-S */
		if (tp.delsys == SYS_DVBS)
		{
			tp.pilot = PILOT_OFF;
			tp.modulation = QPSK;
			tp.rolloff = ROLLOFF_35;
		}
			
		bool accept = false;

		switch (tp.polarisation)
		{
			case 0:
			case 1: accept = (lnb.polarisation == 'L');
				break;
			case 2:
			case 3: accept = (lnb.polarisation == 'C');
			default:break;
		}
		
		if (accept && lnb.frequency)
		{
			int fdiff =
				tp.frequency/1000 - lnb.frequency;

			if (fdiff < -2 || fdiff > +2)
				accept = false;
		}

		if (accept)
			tp_list.push_back(tp);

		tp_num++;
	}
	depth++;
}

namespace
{
	void
	end(void *data, const XML_Char *el __attribute__((unused)))
	{
		((Satellite *)data)->end();
	}
}

void
Satellite::end(void)
{
	depth--;
}

/*
 * Scan XML file
 */
void
Satellite::xmlscan(void)
{
	if (access(xml, R_OK) != 0)
		throw "Can't access xml file";

	ifstream ifs(xml, ios::in | ios::binary);
	if (!ifs.is_open())
		throw "Can't open xml file";

	XML_Parser xmlparser = XML_ParserCreate(NULL);
	if (xmlparser == NULL)
	{
		ifs.close();
		throw "Can't allocate memory for xml parser";
	}

	XML_ParserReset(xmlparser, 0);
	XML_SetUserData(xmlparser, this);
	XML_SetElementHandler(xmlparser, ::start, ::end);

	char cbuf[io::BUFLEN];
	int done = 0;

	while (!done)
	{
		ifs.getline(cbuf, io::BUFLEN);
		if (ifs.eof())
			break;

		if (!XML_Parse(xmlparser, cbuf, strlen(cbuf), done))
		{
			XML_ParserFree(xmlparser);
			ifs.close();
			throw "Can't parse xml";
		}
	}

	XML_ParserFree(xmlparser);
	ifs.close();

	tp_list.sort();
}

/*
 * Scan satellite
 */
void
Satellite::scan(void)
{
	dprintf(20, "frequency = %d\n", lnb.frequency);

	xmlscan();
	if (tp_num == 0)
	{
		cerr << "Satellite not found\n";
		return;
	}
	if (tp_list.size() == 0)
	{
		cerr << "Transponder not found\n";
		return;
	}

	/* Get database connection */
	Sql *db = Sql::getHandle();

	/* Add or update satellite */
	db->addSatellite(this);

	io::outsat(this);

	for (list<Transponder>::iterator ti = tp_list.begin(); ti != tp_list.end(); ++ti)
	{
		Transponder *tp = &(*ti);
		try
		{
			lnb.tune(tp);
		}
		catch (const char *e1)
		{
			io::outkos(this, tp, (string)e1);
			continue;
		}
		catch (string e2)
		{
			io::outkos(this, tp, e2);
			continue;
		}
		if (!lnb.tuned)
		{
			io::outkos(this, tp, "Can't tune");
			continue;
		}
		if (tp->readpat(this))
		{
			tp->readpmt(this);
			tp->readsdt(this);
			tp->readnit(this, Transponder::ANIT);
		}
		sort(tp->service.begin(), tp->service.end());

		/* Add or update transponder and services */
		db->addTransponderAndServices(this, tp);

		/* Output diagnostic details */
		io::outkos(this, tp, "");
	}
}
