/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"
#include "transponder.h"
#include "service.h"

/*
 * Empty service ready for population with scan
 */
Service::Service(__u16 sid, __u16 pmt_pid)
{
	refcount = 1;
	this->sid = sid;
	this->pmt_pid = pmt_pid;
	pcr_pid = 0;
	text = 0;
	video = 0;
	audio.push_back(0);
	type = Service::UNKNOWN;
	typestr = "? ";
	serviceid = -1;
	transponderid = -1;
	name = "";
	lcn = -1;
}

/*
 * ETSI EN 300 468 V1.7.1
 *
 * Table 79: Service type coding
 */
void
Service::settype(__u8 type)
{
	this->type = type;

	/* SDTV */
	if (type == 0x01 || type == 0x16)
		typestr = "SD";

	/* HDTV */
	else if (type == 0x11 || type == 0x19)
		typestr = "HD";

	/* RADIO */
	else if (type == 0x02 || type == 0x0a)
		typestr = "RA";

	/* TEXT */
	else if (type == 0x03)
		typestr = "TX";

	/* RESERVED/UNKNOWN */
	else if (type == 0x00 ||
		(type >= 0x07 && type <= 0x09) ||
		(type >= 0x1c && type <= 0x7f) ||
		type == Service::UNKNOWN)
		typestr = "? ";

	/* NEAR VIDEO ON DEMAND */
	else if (type == 0x04 || type == 0x05 ||
		(type >= 0x17 && type <= 0x18) ||
		(type >= 0x1a && type <= 0x1b))
		typestr = "VD";

	/* USER DEFINED */
	else if (type >= 0x80 && type <= 0xfe)
		typestr = "UD";

	/* MOSAIC */
	else if (type == 0x06 || type == 0x0b)
		typestr = "MO";

	/* DATA/UNKNOWN */
	else
		typestr = "DA";
}

/*
 * Return true when not interested in this service
 */
bool
Service::filter(unsigned int fta, unsigned int typemask)
{
	if (fta && caid.size())
		return true;
	if (typemask)
	{
		unsigned int type = 0;

		if (typestr == "RA")
		{
			if (audio[0])
				type |= RADIO;
		}
		else if (typestr == "SD" || typestr == "HD")
		{
			if (audio[0] && video)
				type |= TV;
		}
		if (typestr == "DA")
		{
			type |= DATA;
		}
		if ((type & typemask) == 0)
			return true;
	}
	return false;
}

/*
 * http://www.etherguidesystems.com/Help/SDOs/MPEG/semantics/mpeg-2/stream_type.aspx
 *
 * MPEG-2 stream_type
 */
void
Service::addstream(int type, int pid)
{
	switch (type)
	{
		/* Video */
		case 0x01:      /* MPEG1 */
		case 0x02:      /* MPEG2 */
		case 0x1b:      /* MPEG4 */
			dprintf(20, "STREAM TYPE (VIDEO) = 0x%02x (%d)\n", type, pid);
			video = pid;
			break;

		/* Audio */
		case 0x03:      /* MPEG1 */
		case 0x04:      /* MPEG2 */
		case 0x0f:      /* AAC   */
		case 0x1c:      /* MPEG4 */
			if (audio[0] == 0)
				audio[0] = pid;
			else
				audio.push_back(pid);
			dprintf(20, "STREAM TYPE (AUDIO) = 0x%02x (%d)\n", type, pid);
			break;

		default:
			dprintf(20, "UNHANDLED STREAM TYPE = 0x%02x (%d)\n", type, pid);
	}
}

/*
 * ETSI EN 300 468 V1.7.1
 *
 * Table 12: Descriptor identification
 */
void
Service::addtag(unsigned char tableid, unsigned char *cbuf, int ix, int tag_id, int tag_len, int pid)
{
	dprintf(20, "tableid = 0x%02x, cbuf = %s, ix = %d, tag_id = 0x%02x, tag_len = 0x%02x, pid = %d\n",
		tableid, (cbuf == NULL) ? "(null)" : "(...)" , ix, tag_id, tag_len, pid);

	if (tableid == Transponder::PMT)
	{
		switch (tag_id)
		{
/* http://www.etherguidesystems.com/Help/SDOs/mpeg/semantics/mpeg-2/descriptors/CA_descriptor.aspx */
			case 0x09:	dprintf(20, "CONDITIONAL ACCESS TAG = 0x%02x TAG LEN = %d\n", tag_id, tag_len);
					if (tag_len >= 4)
					{
						__u16 id  = ID_16(cbuf, ix);
						caid.push_back(id);
						__u16 capid = ID_13(cbuf, ix + 2);
						dprintf(20, "PID = %d CAID = 0x%04x CAPID = %d\n", pid, id, capid);
					}
					break;

/* http://www.etherguidesystems.com/Help/SDOs/mpeg/semantics/mpeg-2/descriptors/ISO_639_language_descriptor.aspx */
			case 0x0a:	dprintf(20, "ISO_639 LANG TAG = 0x%02x TAG LEN = %d\n", tag_id, tag_len);
					if (tag_len >= 4)
					{
						struct language l;
						l.pid = pid;
						for (unsigned int i = 0; i < 3; i++)
						{
							char c = tolower(cbuf[ix + i]);
							if (c >= 'a' && c <= 'z')
								l.language += c;
						}
						l.type = cbuf[ix + 3];
						dprintf(20, "PID = %d LANG = %s TYPE = %d\n",
							l.pid, (l.language).c_str(), l.type);
						lang.push_back(l);
					}
					break;

			case 0x52:	dprintf(20, "STREAM IDENTIFIER TAG = 0x%02x\n", cbuf[ix]);
					break;

			case 0x56:	dprintf(20, "TEXT TAG = 0x%02x PID = %d (0x%04x)\n", tag_id, pid, pid);
					text = pid;
					break;

			case 0x6a:	dprintf(20, "AC-3 AUDIO TAG = 0x%02x PID = %d (0x%04x)\n", tag_id, pid, pid);
					if (audio[0] == 0)
						audio[0] = pid;
					else
						audio.push_back(pid);
					break;

			default:	if (tag_id < 0x80)
						dprintf(20, "UNHANDLED TAG = 0x%02x\n", tag_id);
					break;
		}
	}
	else if (tableid == Transponder::SDT)
	{
		if (tag_id == 0x48)
		{
			unsigned char name_length;
		
			dprintf(20, "SERVICE TAG = 0x%02x\n", tag_id);

			settype(cbuf[ix++]);

			name_length = cbuf[ix++];
			if (name_length)
			{
				provider = cstring::annexa(&cbuf[ix], name_length);
				dprintf(20, "PROVIDER NAME = %s\n", provider.c_str());
				ix += name_length;
			}

			name_length = cbuf[ix++];
			if (name_length)
			{
				service = cstring::annexa(&cbuf[ix], name_length);
				dprintf(20, "SERVICE NAME = %s\n", service.c_str());
				ix += name_length;
			}
		}
		else if (tag_id < 0x80)
			dprintf(20, "UNHANDLED TAG = 0x%02x\n", tag_id);
	}
	else
		dprintf(20, "UNHANDLED TABLE = 0x%02x\n", tableid);
}

void
Service::readdesc(unsigned char tableid, unsigned char *cbuf, int info_length, int pid)
{
	dprintf(20, "tableid = %d, cbuf = %s, info_length = %d, pid = %d\n",
		tableid, (cbuf == NULL) ? "(null)" : "(...)", info_length, pid);

	int ix = 0;
	while (ix < info_length)
	{
		unsigned char tag_id = cbuf[ix++];
		unsigned char tag_len = cbuf[ix++];
		
		if (tag_len)
		{
			addtag(tableid, cbuf, ix, tag_id, tag_len, pid);
			ix += tag_len;
		}
	}
}

/*
 * Serialise caids
 */
string
Service::serialiseCaid(void)
{
	string ids = "";

	if (caid.size())
	{
		sort(caid.begin(), caid.end());
		vector<__u16>::iterator e = unique(caid.begin(), caid.end());

		for (vector<__u16>::iterator i = caid.begin(); i < e; i++)
		{
			stringstream id;
			id << setfill('0') << hex << setw(4) << *i;
			if (ids == "")
				ids = id.str();
			else
				ids = ids + "," + id.str();
		}
	}
	else
		ids = "0";

	return ids;
}

/*
 * Unserialise caids
 */
void
Service::unserialiseCaid(string s)
{
	caid.clear();

	if (s.find_first_of(',') == string::npos)
	{
		if (s != "0")
			caid.push_back(util::toInt(s, 16));
	}
	else
	{
		vector<string> c = util::split(s, ',', 0);
		for (vector<string>::iterator i = c.begin(); i < c.end(); *i++)
			caid.push_back(util::toInt(*i, 16));
	}
}

/*
 * Serialise audio
 */
string
Service::serialiseAudio(void)
{
	string pids = "";

	for (vector<__u16>::iterator i = audio.begin(); i < audio.end(); i++)
	{
		/* Default to no langauge */
		string thispid = util::toString(*i);

		/* Determine if language is known for this pid */
		for (vector<struct language>::iterator l = lang.begin(); l < lang.end(); l++)
		{
			if (*i == l->pid)
			{
				thispid = thispid + "=" + l->language;
				break;
			}
		}
		if (pids == "")
			pids = thispid;
		else
			pids = pids + "," + thispid;
	}
	return pids;
}

/*
 * unserialise audio
 */
void
Service::unserialiseAudio(string s)
{
	audio.clear();

	if (s.find_first_of(',') == string::npos)
	{
		if (s.find_first_of('=') == string::npos)
		{
			audio.push_back(util::toInt(s));
		}
		else
		{
			vector<string> s1 = util::split(s, '=', 0);
			audio.push_back(util::toInt(s1[0]));

			struct language l;
			l.pid = util::toInt(s1[0]);
			l.language = s1[1];
			l.type = '\0'; /* XXX */
			lang.push_back(l);
		}
	}
	else
	{
		vector<string> s1 = util::split(s, ',', 0);
		for (vector<string>::iterator i = s1.begin(); i < s1.end(); i++)
		{
			if (i->find_first_of('=') == string::npos)
			{
				audio.push_back(util::toInt(*i));
			}
			else
			{
				vector<string> s2 = util::split(*i, '=', 0);
				audio.push_back(util::toInt(s2[0]));

				struct language l;
				l.pid = util::toInt(s2[0]);
				l.language = s2[1];
				l.type = '\0'; /* XXX */
				lang.push_back(l);
			}
		}
	}
}
