/*
 * Copyright (c) 2004-2008 Darron M Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "dvbcompat.h"
#include "cstring.h"
#include "util.h"
#include "dotconf.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"

#include "matrix.h"
#include "sql.h"

extern "C" {
#include <mysql/mysql.h>
};
#include "mysql5.h"

#include "sqlite.h"

extern "C" {
#include <sqlite.h>
};
#include "sqlite2.h"

extern "C" {
#include <sqlite3.h>
};
#include "sqlite3.h"

/*
 * sql interface class
 */

/* Database handle */
Sql *Sql::handle = NULL;

Sql::Sql(void)
{
}

Sql::~Sql(void)
{
}

/*
 * Return SQL database handle
 */
Sql *
Sql::getHandle(bool initialise)
{
	dprintf(20, "HANDLE = %p\n", handle);

	if (handle == NULL)
	{
		Dotconf config;
		if (strcasecmp(config.dbtype.c_str(), "MySQL5") == 0)
		{
			handle = new Mysql5(config.dbhost.c_str(),
					config.dbuser.c_str(),
					config.dbpass.c_str(),
					config.dbname.c_str(),
					initialise);
		}
		else if (strcasecmp(config.dbtype.c_str(), "SQLite2") == 0)
		{
			handle = new Sqlite2(config.dbname.c_str(), initialise);
		}
		else if (strcasecmp(config.dbtype.c_str(), "SQLite3") == 0)
		{
			handle = new Sqlite3(config.dbname.c_str(), initialise);
		}
		else
		{
			string error = "Unknown database type [MySQL5 | SQLite2 | SQLite3] : " +
				config.dbtype;
			throw error;
		}
	}
	else
		handle->reconnect();

	return handle;
}

/*
 * Perform SQL commands in text file
 */
void
Sql::file(const char *file)
{
	if (access(file, R_OK) != 0)
	{
		string error = "error: cannot access file [" + (string)file + "]";
		throw error;
	}

	ifstream ifs(file, ios::in | ios::binary);

	if (!ifs.is_open())
	{
		string error = "error: cannot open file [" + (string)file + "]";
		throw error;
	}

	char cbuf[cstring::STRLEN];
	string command = "";
	while (true)
	{
		ifs.getline(cbuf, cstring::STRLEN);
		if (ifs.eof())
			break;

		if (cbuf[0] && cbuf[0] != '-')
		{
			command += cbuf;
			if (cbuf[strlen(cbuf) - 1] == ';')
			{
				SQLcommand(command.c_str());
				command = "";
			}
			else
				command += "\n";
		}
	}
	ifs.close();
}

/*
 * Perform a query and return 0,0 as int else error flag
 */
int
Sql::getId(const char *query)
{
	SqlMatrix rs;
	int count = SQLquery(query, &rs);

	if (count == 0)
		return -1;

	return util::toInt(rs.get(0, 0));
}
int
Sql::getId(string query)
{
	return getId(query.c_str());
}
int
Sql::getId(stringstream &query)
{
	return getId(query.str().c_str());
}

/*
 * Get satellite id using unique position
 */
int
Sql::getSatelliteId(class Satellite *sat)
{
	stringstream query;

	query << "SELECT satelliteid FROM satellite WHERE positionid = ";
	query << "'" << sat->lnb.position << "'";
	
	return getId(query);
}

/*
 * Add a satellite
 */
int
Sql::addSatellite(class Satellite *sat)
{
	sat->satelliteid = getSatelliteId(sat);

	if (sat->satelliteid < 0)
	{
		stringstream query;

		query << "INSERT INTO satellite (name,positionid,adapter,port,";
		query << SQLquotefield("repeat") << ",";
		query << "burst,low,high,switch,polarisation,timeout,timestamp) VALUES (";
		query << "'" << SQLescape(sat->name)	<< "',";
		query << "'" << sat->lnb.position	<< "',";
		query << "'" << sat->lnb.phy_adapter	<< "',";
		query << "'" << (int)sat->lnb.dsec_port	<< "',";
		query << "'" << sat->lnb.dsec_repeat	<< "',";
		query << "'" << sat->lnb.dsec_burst	<< "',";
		query << "'" << sat->lnb.low		<< "',";
		query << "'" << sat->lnb.high		<< "',";
		query << "'" << sat->lnb.threshold	<< "',";
		query << "'" << sat->lnb.polarisation	<< "',";
		query << "'" << sat->lnb.tuning_delay	<< "',";
		query << "'" << util::timestamp()	<< "')";

		SQLcommand(query);
	
		sat->satelliteid = getSatelliteId(sat);
	}
	else
		updateSatellite(sat);

	return sat->satelliteid;
}

/*
 * Update satellite
 */
int
Sql::updateSatellite(class Satellite *sat)
{
	if (sat->satelliteid >= 0)
	{
		stringstream query;

		query << "UPDATE satellite SET ";
		query << "name = '"		<< SQLescape(sat->name) << "',";
		query << "positionid = '"	<< sat->lnb.position	<< "',";
		query << "adapter = '"		<< sat->lnb.phy_adapter << "',";
		query << "port = '"		<< (int)sat->lnb.dsec_port<< "',";
		query << SQLquotefield("repeat") << " = '"
						<< sat->lnb.dsec_repeat	<< "',";
		query << "burst = '"		<< sat->lnb.dsec_burst	<< "',";
		query << "low = '"		<< sat->lnb.low		<< "',";
		query << "high = '"		<< sat->lnb.high	<< "',";
		query << "switch = '"		<< sat->lnb.threshold	<< "',";
		query << "polarisation = '"	<< sat->lnb.polarisation<< "',";
		query << "timeout = '"		<< sat->lnb.tuning_delay<< "',";
		query << "timestamp = '"	<< util::timestamp()	<< "' ";

		query << "WHERE satelliteid = '" << sat->satelliteid << "'";

		SQLcommand(query);

		return sat->satelliteid;
	}

	return -1;
}

/*
 * Get transponder id using unique index
 */
int
Sql::getTransponderId(class Transponder *tp)
{
	stringstream query;

	query << "SELECT transponderid FROM transponder WHERE ";
	query << "satelliteid = '"  << tp->satelliteid << "' AND ";
	query << "frequency = '"    << tp->frequency   << "' AND ";
	query << "polarisation = '" << dvbcompat::polarisation2cstr(tp->polarisation) << "'";
	
	return getId(query);
}

/*
 * Add a transponder
 */
int
Sql::addTransponder(int satelliteid, class Transponder *tp)
{
	tp->satelliteid = satelliteid;
	tp->transponderid = getTransponderId(tp);

	if (tp->transponderid < 0)
	{
		stringstream query;

		query << "INSERT INTO transponder (satelliteid,frequency,polarisation,";
		query << "delsys,modulation,symbolrate,fec,rolloff,pilot,timestamp) VALUES (";
		query << "'" <<	satelliteid					<< "',";
		query << "'" <<	tp->frequency					<< "',";
		query << "'" <<	dvbcompat::polarisation2cstr(tp->polarisation)	<< "',";
		query << "'" <<	dvbcompat::delsyst2cstr(tp->delsys)		<< "',";
		query << "'" <<	dvbcompat::modulationt2cstr(tp->modulation)	<< "',";
		query << "'" <<	tp->symbol_rate					<< "',";
		query << "'" <<	dvbcompat::coderatet2cstr(tp->coderate)		<< "',";
		query << "'" <<	dvbcompat::rollofft2cstr(tp->rolloff)		<< "',";
		query << "'" <<	dvbcompat::pilott2cstr(tp->pilot)		<< "',";
		query << "'" <<	util::timestamp()				<< "')";

		SQLcommand(query);
	
		tp->transponderid = getTransponderId(tp);
		
		return updateTransponderDefaultUd(tp);
	}

	return updateTransponder(tp);
}

/*
 * Update transponder
 */
int
Sql::updateTransponder(class Transponder *tp)
{
	if (tp->transponderid >= 0)
	{
		stringstream query;

		query << "UPDATE transponder SET ";
		query << "delsys = '" << dvbcompat::delsyst2cstr(tp->delsys) << "',";
		query << "modulation = '" << dvbcompat::modulationt2cstr(tp->modulation) << "',";
		query << "symbolrate = '" << tp->symbol_rate << "',";
		query << "fec = '" << dvbcompat::coderatet2cstr(tp->coderate) << "',";
		query << "rolloff = '" << dvbcompat::rollofft2cstr(tp->rolloff) << "',";
		query << "pilot = '" << dvbcompat::pilott2cstr(tp->pilot) << "',";
		query << "timestamp = '" << util::timestamp() << "' ";
		query << "WHERE transponderid = '" << tp->transponderid << "'";

		SQLcommand(query);

		return updateTransponderDefaultUd(tp);
	}

	return -1;
}

/*
 * Update transponder default user-defined name
 *
 * name is only updated if NULL, else left as-is
 */
int
Sql::updateTransponderDefaultUd(class Transponder *tp)
{
	if (tp->transponderid >= 0)
	{
		/* default UD name */
		string newname = SQLescape(tp->name);
		if (newname == "")
			newname = "[" + util::toString(tp->frequency / 1000) +
				dvbcompat::polarisation2cstr(tp->polarisation) + "]";

		stringstream query;

		query << "UPDATE transponder SET ";
		query << "name = '" << newname << "',";
		query << "timestamp = '" << util::timestamp() << "' ";
		query << "WHERE transponderid = '" << tp->transponderid << "' AND ";
		query << SQLnull("name");

		SQLcommand(query);

		return tp->transponderid;
	}

	return -1;
}

/*
 * Update transponder from descriptors
 *
 * Network ID, Network name, Original network ID, transport ID
 * and roll-off are learnt from the stream
 */
int
Sql::updateTransponderDesc(class Transponder *tp)
{
	if (tp->transponderid >= 0)
	{
		stringstream query;

		query << "UPDATE transponder SET ";
		query << "network = '" << SQLescape(tp->network) << "',";
		query << "nit = '" << tp->nit << "',";
		query << "nid = '" << tp->nid << "',";
		query << "onid = '" << tp->onid << "',";
		query << "tid = '" << tp->tid << "',";
		query << "rolloff = '" << dvbcompat::rollofft2cstr(tp->rolloff) << "',";
		query << "timestamp = '" << util::timestamp() << "' ";
		query << "WHERE transponderid = '" << tp->transponderid << "'";
		
		SQLcommand(query);

		return tp->transponderid;
	}

	return -1;
}

/*
 * Add or update tranponder and services
 */
int
Sql::addTransponderAndServices(class Satellite *sat, class Transponder *tp)
{
	/* Add or update transponder with info learnt from the input */
	addTransponder(sat->satelliteid, tp);

	/* Add or update transponder with info learnt from descriptors */
	updateTransponderDesc(tp);

	/* Add or update services on this transponder */
	if (tp->service.size())
	{
		for (unsigned int srvix = 0; srvix < tp->service.size(); srvix++)
			addService(tp->transponderid, &tp->service[srvix]);

		return tp->service.size();
	}

	return -1;
}

/*
 * Get service id using unique index
 */
int
Sql::getServiceId(class Service *service)
{
	stringstream query;

	query << "SELECT serviceid FROM service WHERE ";
	query << "transponderid = '" << service->transponderid << "' AND ";
	query << "sid = '" << service->sid << "'";
	
	return getId(query);
}

/*
 * Add a service
 */
int
Sql::addService(int transponderid, class Service *service)
{
	service->transponderid = transponderid;
	service->serviceid = getServiceId(service);

	if (service->serviceid < 0)
	{
		stringstream query;

		query << "INSERT INTO service (transponderid,typ,service,provider,";
		query << "sid,vpid,apid,pmt,pcr,txt,audio,caid,timestamp) VALUES (";
		query << "'" <<	service->transponderid << "',";
		query << "'" <<	(int)service->type << "',";
		query << "'" <<	SQLescape(service->service) << "',";
		query << "'" <<	SQLescape(service->provider) << "',";
		query << "'" <<	service->sid << "',";
		query << "'" <<	service->video << "',";
		query << "'" <<	service->audio[0] << "',";
		query << "'" <<	service->pmt_pid << "',";
		query << "'" <<	service->pcr_pid << "',";
		query << "'" <<	service->text << "',";
		query << "'" <<	SQLescape(service->serialiseAudio()) << "',";
		query << "'" <<	SQLescape(service->serialiseCaid()) << "',";
		query << "'" <<	util::timestamp() << "')";;

		SQLcommand(query);
	
		service->serviceid = getServiceId(service);
		
		return updateServiceDefaultUd(service);
	}

	return updateService(service);
}

/*
 * Update on-air service
 */
int
Sql::updateService(class Service *service)
{
	if ((service->serviceid < 0) ||
		(service->pmt_pid == 0) ||
			(service->audio[0] == 0 && service->typestr == "RA") ||
				((service->audio[0] == 0 || service->video == 0) &&
				(service->typestr == "SD" || service->typestr == "HD")))
		return -1;
	
	stringstream query;

	query << "UPDATE service SET ";
	query << "typ = '" << (int)service->type << "',";
	query << "service = '" << SQLescape(service->service) << "',";
	query << "provider = '" << SQLescape(service->provider) << "',";
	query << "vpid = '" << service->video << "',";
	query << "apid = '" << service->audio[0] << "',";
	query << "pmt = '" << service->pmt_pid << "',";
	query << "pcr = '" << service->pcr_pid << "',";
	query << "txt  = '" << service->text << "',";
	query << "audio = '" << SQLescape(service->serialiseAudio()) << "',";
	query << "caid = '" << SQLescape(service->serialiseCaid()) << "',";
	query << "timestamp = '" << util::timestamp() << "' ";
	query << "WHERE serviceid = '" << service->serviceid << "'";

	SQLcommand(query);

	return updateServiceDefaultUd(service);
}

/*
 * Update service default user-defined name, lcn
 *
 * name and lcn are only updated if NULL, else left as-is
 */
int
Sql::updateServiceDefaultUd(class Service *service)
{
	if (service->serviceid >= 0)
	{
		/* default UD name */
		string newname = SQLescape(service->service);
		if (newname == "")
			newname = "[" + util::toString(service->sid) + "]";

		stringstream query;

		query << "UPDATE service SET ";
		query << "name = '" << newname << "', ";
		query << "timestamp = '" << util::timestamp() << "' ";
		query << "WHERE serviceid = '" << service->serviceid << "' AND ";
		query << SQLnull("name");

		SQLcommand(query);

		/* default UD LCN */
		query.str("");
		query << "UPDATE service SET ";
		query << "lcn = '" << service->serviceid << "',";
		query << "timestamp = '" << util::timestamp() << "' ";
		query << "WHERE serviceid = '" << service->serviceid << "' AND ";
		query << SQLnull("lcn");
		
		SQLcommand(query);

		return service->serviceid;
	}

	return -1;
}

/*
 * Get satellites at all positions
 */
int
Sql::getSatellites(SqlMatrix *rs)
{
	return SQLquery("SELECT * FROM satellite ORDER BY satelliteid", rs);
}

/*
 * Get transponders on satellite
 */
int
Sql::getTransponders(int satelliteid, SqlMatrix *rs)
{
	stringstream query;
	
	query << "SELECT * FROM transponder WHERE ";
	query << "satelliteid = '" << satelliteid << "' ";
	query << "ORDER BY transponderid";

	return SQLquery(query, rs);
}

/*
 * Get services on transponder
 */
int
Sql::getServices(int transponderid, SqlMatrix *rs)
{
	stringstream query;

	query << "SELECT * FROM service WHERE ";
	query << "transponderid = '" << transponderid << "' ";
	query << "ORDER BY serviceid";

	return SQLquery(query, rs);
}

/*
 * Create a new service instance from an sql result set
 */
class Service 
Sql::newService(SqlMatrix *rs, unsigned int row)
{
	class Service service;

	service.serviceid = rs->asInt("serviceid", row);
	service.transponderid = rs->asInt("transponderid", row);
	service.name = rs->asString("name", row);
	service.lcn = rs->asInt("lcn", row);
	service.type = rs->asInt("typ", row); service.settype(service.type);
	service.service = rs->asString("service", row);
	service.provider = rs->asString("provider", row);
	service.sid = rs->asInt("sid", row);
	service.video = rs->asInt("vpid", row);
	service.pmt_pid = rs->asInt("pmt", row);
	service.pcr_pid = rs->asInt("pcr", row);
	service.text = rs->asInt("txt", row);
	service.unserialiseAudio(rs->asString("audio", row));
	service.unserialiseCaid(rs->asString("caid", row));

	return service;
}

/*
 * Create a new transponder instance from an sql result set
 */
class Transponder
Sql::newTransponder(SqlMatrix *rs, unsigned int row)
{
	class Transponder tp;

	tp.transponderid = rs->asInt("transponderid", row);
	tp.satelliteid = rs->asInt("satelliteid", row);
	tp.name = rs->asString("name", row);
	tp.frequency = rs->asInt("frequency", row);
	tp.polarisation = dvbcompat::polarisation2uint(rs->asChar("polarisation", row));
	tp.voltage = dvbcompat::polarisation2voltaget(tp.polarisation);
	tp.symbol_rate = rs->asInt("symbolrate", row);
	tp.delsys = dvbcompat::str2delsyst(rs->asString("delsys", row));
	tp.modulation = dvbcompat::str2modulationt(rs->asString("modulation", row));
	tp.coderate = dvbcompat::str2coderatet(rs->asString("fec", row));
	tp.rolloff = dvbcompat::str2rollofft(rs->asString("rolloff", row));
	tp.pilot = dvbcompat::str2pilott(rs->asString("pilot", row));
	tp.nit = rs->asInt("nit", row);
	tp.network = rs->asString("network", row);
	tp.nid = rs->asInt("nid", row);
	tp.onid = rs->asInt("onid", row);
	tp.tid = rs->asInt("tid", row);

	return tp;
}

/*
 * Create a new satellite instance from an sql result set
 */
class Satellite
Sql::newSatellite(SqlMatrix *rs, unsigned int row)
{
	class Satellite sat;

	sat.satelliteid = rs->asInt("satelliteid", row);
	sat.lnb.position = rs->asInt("positionid", row);
	sat.name = rs->asString("name", row);
	sat.lnb.phy_adapter = rs->asInt("adapter", row);
	sat.lnb.dsec_port = rs->asInt("port", row);
	sat.lnb.dsec_repeat = rs->asInt("repeat", row);
	sat.lnb.dsec_burst = rs->asInt("burst", row);
	sat.lnb.low = rs->asInt("low", row);
	sat.lnb.high = rs->asInt("high", row);
	sat.lnb.threshold = rs->asInt("switch", row);
	sat.lnb.polarisation = rs->asChar("polarisation", row);
	sat.lnb.tuning_delay = rs->asInt("timeout", row);

	return sat;
}

/*
 * unserialise satellite from database
 */
class Satellite
Sql::getSatellite(SqlMatrix *rs, unsigned int row)
{
	class Satellite sat = newSatellite(rs, row);

	SqlMatrix rsTp;
	sat.tp_num = getTransponders(sat.satelliteid, &rsTp);

	for (unsigned tpix = 0 ; tpix < sat.tp_num ; tpix++)
	{
		class Transponder tp = newTransponder(&rsTp, tpix);

		SqlMatrix rsSrv;

		int srvcnt = getServices(tp.transponderid, &rsSrv);
		for (int srvix = 0; srvix < srvcnt ; srvix++)
			tp.service.push_back(newService(&rsSrv, srvix));
		
		sat.tp_list.push_back(tp);
	}
	return sat;
}
