/*
 * Copyright (c) 2004-2008 Darron M Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"

#include "matrix.h"
#include "sql.h"

#include "sqlite.h"

extern "C" {
#include <sqlite.h>
};
#include "sqlite2.h"

/*
 * sqlite2 interface class
 */

Sqlite2::Sqlite2(const char *database_file, bool initialise, const char *schema_file)
{
	/* if init required, prepare file */
	if (initialise)
	{
		if (::access(database_file, F_OK) == 0)
		{
			if (::unlink(database_file) != 0)
				throw "Can't re-init database";
		}
	}
	
	/* open existant or create new database */
	db = sqlite_open(database_file, 0, NULL);
	if( db == 0 )
		throw "Can't create or open database";

	/* if database didn't exist and was created, initialise from schema */
	if (initialise)
		file(schema_file);
}

Sqlite2::~Sqlite2(void)
{
	sqlite_close(db);
}

/*
 * perform a query that has no results
 */
void
Sqlite2::SQLcommand(const char *query)
{
	int rc, w = 0;

	while(true)
	{
		rc = sqlite_exec(db, query, NULL, NULL, NULL);
		if( rc == SQLITE_BUSY || rc == SQLITE_LOCKED )
			SQLwait(&w);
		else
			break;
	}
	if (rc != SQLITE_OK)
	{
		string error = "error: SQLcommand in '" + (string)query + "'";
		throw error;
	}
}
void
Sqlite2::SQLcommand(string query)
{
	SQLcommand(query.c_str());
}
void
Sqlite2::SQLcommand(stringstream &query)
{
	SQLcommand(query.str().c_str());
}

/*
 * peform query and return count
 *
 * nb. query must start as SELECT COUNT(*) FROM ...
 */
int
Sqlite2::SQLcount(const char *query)
{
	int rc, w = 0, nrow = 0, ncol = 0;
	char **resultset;

	while(true)
	{
		rc = sqlite_get_table(db, query, &resultset, &nrow, &ncol, NULL);
		if (rc == SQLITE_BUSY || rc == SQLITE_LOCKED)
			SQLwait(&w);
		else
			break;
	}
	if (rc != SQLITE_OK)
		throw "error: SQLcount";

	int count=0;
	if (nrow == 1 && ncol == 1)
		count= atoi(resultset[1]);
	sqlite_free_table(resultset);

	return count;
}
int
Sqlite2::SQLcount(string query)
{
	return SQLcount(query.c_str());
}
int
Sqlite2::SQLcount(stringstream &query)
{
	return SQLcount(query.str().c_str());
}

/*
 * peform query and return number of rows and populate a result matrix
 */
int
Sqlite2::SQLquery(const char *query, SqlMatrix *matrix)
{
	int rc, w = 0, nrow = 0, ncol = 0;
	char **resultset;

	while(true)
	{
		rc = sqlite_get_table(db, query, &resultset, &nrow, &ncol, NULL);
		if (rc == SQLITE_BUSY || rc == SQLITE_LOCKED)
			SQLwait(&w);
		else
			break;
	}
	if (rc != SQLITE_OK)
		throw "error: SQLquery";
	if (nrow == 0)
		return 0;

	matrix->resize(ncol, nrow);

	for (int x = 0; x < ncol; x++)
	{
		string fieldname = resultset[x];
		int pos = 1 + fieldname.find_last_of('.');

		string newfieldname; /* FIXME substr */
		for (unsigned int i = pos; i < fieldname.length(); i++)
			newfieldname += fieldname[i];

		matrix->addFieldName(newfieldname);
	}

	int index = ncol; /* skip field labels */
	for(int y = 0; y < nrow; y++)
	{
		for(int x = 0; x < ncol; x++)
		{
			if (resultset[index] == NULL)
				matrix->set(x, y, (string)"(null)");
			else
				matrix->set(x, y, resultset[index]);
			index++;
		}
	}

	sqlite_free_table(resultset);
	
	return nrow;
}
int
Sqlite2::SQLquery(string query, SqlMatrix *matrix)
{
	return SQLquery(query.c_str(), matrix);
}
int
Sqlite2::SQLquery(stringstream &query, SqlMatrix *matrix)
{
	return SQLquery(query.str().c_str(), matrix);
}
