/*
 * Copyright (c) 2004-2008 Darron M Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"

#include "matrix.h"
#include "sql.h"

#include "sqlite.h"

extern "C" {
#include <sqlite3.h>
};
#include "sqlite3.h"

/*
 * sqlite3 interface class
 */

Sqlite3::Sqlite3(const char *database_file, bool initialise, const char *schema_file)
{
	/* if init required, prepare file */
	if (initialise)
	{
		if (::access(database_file, F_OK) == 0)
		{
			if (::unlink(database_file) != 0)
				throw "Can't re-init database";
		}
	}

	/* open existant or create new database */
	int rc = sqlite3_open(database_file, &db);
	if (rc != SQLITE_OK)
		throw sqlite3_errmsg(db);

	/* if database didn't exist and was created, initialise from schema */
	if (initialise)
		file(schema_file);
}

Sqlite3::~Sqlite3(void)
{
	sqlite3_close(db);
}

/*
 * perform a query that has no results
 */
void
Sqlite3::SQLcommand(const char *query)
{
	int rc, w = 0;

	while (true)
	{
		rc = sqlite3_exec(db, query, NULL, NULL, NULL);
		if( rc == SQLITE_BUSY || rc == SQLITE_LOCKED )
			SQLwait(&w);
		else
			break;
	}
	if (rc != SQLITE_OK)
	{
		string error = (string)sqlite3_errmsg(db) + " in '" + (string)query + "'";
		throw error;
	}
}
void
Sqlite3::SQLcommand(string query)
{
	SQLcommand(query.c_str());
}
void
Sqlite3::SQLcommand(stringstream &query)
{
	SQLcommand(query.str().c_str());
}

/*
 * peform query and return count
 *
 * nb. query must start as SELECT COUNT(*) FROM ...
 */
int
Sqlite3::SQLcount(const char *query)
{
	int rc, w = 0, nrow = 0, ncol = 0;
	char **resultset;

	while (true)
	{
		rc = sqlite3_get_table(db, query, &resultset, &nrow, &ncol, NULL);
		if (rc == SQLITE_BUSY || rc == SQLITE_LOCKED)
			SQLwait(&w);
		else
			break;
	}
	if (rc != SQLITE_OK)
		throw sqlite3_errmsg(db);

	int count = 0;
	if (nrow == 1 && ncol == 1)
		count = atoi(resultset[1]);
	sqlite3_free_table(resultset);

	return count;
}
int
Sqlite3::SQLcount(string query)
{
	return SQLcount(query.c_str());
}
int
Sqlite3::SQLcount(stringstream &query)
{
	return SQLcount(query.str().c_str());
}

/*
 * peform query and return number of rows and populate a result matrix
 */
int
Sqlite3::SQLquery(const char *query, SqlMatrix *matrix)
{
	int rc, w = 0, nrow = 0, ncol = 0;
	char **resultset;

	while (true)
	{
		rc = sqlite3_get_table(db, query, &resultset, &nrow, &ncol, NULL);
		if (rc == SQLITE_BUSY || rc == SQLITE_LOCKED )
			SQLwait(&w);
		else
			break;
	}
	if (rc != SQLITE_OK)
		throw sqlite3_errmsg(db);
	if (nrow == 0)
		return 0;

	matrix->resize(ncol, nrow);

	for (int x = 0; x < ncol; x++)
		matrix->addFieldName(resultset[x]);

	int index=ncol;	/* skip field labels */
	for (int y = 0; y < nrow; y++)
	{
		for(int x = 0; x < ncol; x++)
		{
			if (resultset[index] == NULL)
				matrix->set(x, y, (string)"(null)");
			else
				matrix->set(x, y, resultset[index]);
			index++;
		}
	}

	sqlite3_free_table(resultset);

	return nrow;
}
int
Sqlite3::SQLquery(string query, SqlMatrix *matrix)
{
	return SQLquery(query.c_str(), matrix);
}
int
Sqlite3::SQLquery(stringstream &query, SqlMatrix *matrix)
{
	return SQLquery(query.str().c_str(), matrix);
}
