/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"
#include "matrix.h"
#include "sql.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"
#include "sqlsat.h"
#include "io.h"

void
usage(char *execname, string errstr, int errnum)
{
	cerr << "USAGE: " << execname << " [OPTIONS] COMMAND [ARGS]\n\n";

	cerr <<	"Output options:\n"
		" -a\t\tOutput all broadcasts.\n"
		" -f\t\tOutput only FTA broadcasts.\n"
		" -r\t\tOutput includes only radio broadcasts.\n"
		" -t\t\tOutput includes only television broadcasts.\n\n"
		" (default)\t-ft\n\n";
	
	cerr << "Commands:\n"
		" INIT\t\tInitialise database as configured in `config'.\n"
		" KOS\t\tDump channels in `KingOfSat' format.\n"
		" SZAP\t\tExport channels.conf in szap-s2 format.\n"
		" VDR\t\tExport channels.conf in vdr format.\n";

	if (errstr != "")
		cerr << "\nERROR:\n " << errstr << ".\n";

	exit(errnum);
}

int
main(int argc, char **argv)
{
	unsigned int fta = 1, filter = Service::TV;
	char c, *execdup, *execname, *command;

	execdup = strdup(argv[0]);
	execname = basename(execdup);

	opterr = 0;
	while ((c = getopt(argc, argv, "afrt")) != -1)
	{
		switch (c)
		{
			case 'a':
				fta = 0;
				filter = Service::TV | Service::RADIO | Service::DATA;
				break;
			case 'f':
				fta = 1;
				filter = Service::TV | Service::RADIO;
				break;
			case 'r':
				filter = Service::RADIO;
				break;
			case 't':
				filter = Service::TV;
				break;
			case '?':
				usage(execname, "", EX_OK);
				break;
			default:
				string err = "Unknown option " + (string)optarg;
				usage(execname, err, EX_DATAERR);
				break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
		usage(execname, "Missing args(s)", EX_DATAERR);

	command = argv[0];

	setlocale(LC_ALL, "");

	try
	{
		/*  Initialise database */
		if (strcasecmp(command, "INIT") == 0)
		{
			if (argc > 1)
				usage(execname, "Too many args(s)", EX_DATAERR);
			
			Sql::getHandle(true);
			cout << "Database initialised.\n";
		}

		/* Output services in KOS, SZAP or VDR format */
		else if ((strcasecmp(command, "KOS") == 0) ||
			(strcasecmp(command, "SZAP") == 0) ||
			(strcasecmp(command, "VDR") == 0))
		{
			if (argc > 1)
				usage(execname, "Too many args(s)", EX_DATAERR);

			Sql *db = Sql::getHandle();
			SqlMatrix rsSat;

			int satcnt = db->getSatellites(&rsSat);
			if (satcnt > 0) for (int satix = 0; satix < satcnt; satix++)
			{
				class Satellite sat = db->getSatellite(&rsSat, satix);
				for (list<Transponder>::iterator ti = sat.tp_list.begin(); ti != sat.tp_list.end(); ++ti)
				{
					if (strcasecmp(command, "KOS") == 0)
					{
						io::outkos(&sat, &(*ti), "");
					}
					else for (vector<class Service>::iterator si = ti->service.begin(); si < ti->service.end(); si++)
					{
						if (si->filter(fta, filter) == false)
						{
							if (strcasecmp(command, "SZAP") == 0)
								io::outszap(&sat, &(*ti), &(*si));
							else
								io::outvdr(&sat, &(*ti), &(*si));
						}
					}
				}
			}
			else
				cout << "No satellites found.\n";
		}

		else
		{
			string error = "Unknown command `" + (string)command + "'";
			usage(execname, error, EX_DATAERR);
		}
	}
	catch (const char *e1)
	{
		cerr << "\nError: " << e1 << endl;
	}
	catch (string e2)
	{
		cerr << "\nError: " << e2 << endl;
	}
	exit(EX_OK);
}
