/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "dvbcompat.h"
#include "cstring.h"
#include "util.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"
#include "io.h"

int
Transponder::read(class Satellite *sat, unsigned char *cbuf, unsigned char tableid, unsigned int section_number, unsigned int *last_section_number)
{
	unsigned int this_section_number;
	int section_length;
	do
	{
		int len = sat->lnb.read(cbuf);
		if (len < 12)
			return 0;

		unsigned char table_id = cbuf[0];
		assert(table_id == tableid);

		section_length = SECTION_LENGTH(cbuf);
		assert(section_length == len);
		
		this_section_number = cbuf[6];
		*last_section_number = cbuf[7];

		dprintf(20, "TABLE ID=0x%02x SECTION LENGTH=%d THIS SECTION #=%d WANTED SECTION #=%d LAST SECTION #=%d\n",
			table_id, section_length - CRC_LENGTH, this_section_number, section_number, *last_section_number);
	}
	while (section_number != this_section_number);

	return section_length - CRC_LENGTH;
}

void
Transponder::addtag(unsigned char tableid, unsigned char *cbuf, int ix, int tag_id, int tag_len)
{
	dprintf(20, "tableid = 0x%02x, cbuf = %s, ix = %d, tag_id = 0x%02x, tag_len = 0x%02x\n",
		tableid, (cbuf == NULL) ? "(null)" : "(...)" , ix, tag_id, tag_len);

	if (tableid == ANIT)
	{
		switch (tag_id)
		{
			case 0x40:
				/* Network name. We may receive multiple but just use the last one */
				{
				network = cstring::annexa(&cbuf[ix], tag_len);
				dprintf(20, "NETWORK TAG = 0x%02x NAME = %s\n", tag_id, network.c_str());
				}
				break;

			case 0x43:
				{
				stringstream ss;

				dprintf(25, "SATELLITE DELIVERY SYSTEM TAG = 0x%02x TAG LEN = %d\n",
					tag_id, tag_len);

				/* Frequency */
				ss << setfill(' ') << hex << (int)cbuf[ix++];
				ss << setfill('0');
				ss << hex << setw(2) << (int)cbuf[ix++];
				ss << hex << setw(2) << (int)cbuf[ix++];
				ss << hex << setw(2) << (int)cbuf[ix++] << '0';
				dprintf(25, "FREQUENCY: %s\n", ss.str().c_str());
				
				/* Satellite position */
				ss.str("");
				ss << setfill(' ') << hex << (int)cbuf[ix++];
				ss << setfill('0') << hex << setw(2) << (int)cbuf[ix++];

				/* East/West position 0=W, 1=E */
				unsigned char p = (cbuf[ix] & 0x80) >> 7;
				if (p == 0) {
					dprintf(25, "POSITION: -%s\n", ss.str().c_str());
				}
				else {
					dprintf(25, "POSITION: %s\n", ss.str().c_str());
				}

				fe_sec_voltage_t     v = dvbcompat::polarisation2voltaget((cbuf[ix] & 0x60) >> 5);
				fe_rolloff_t         r = dvbcompat::rolloff2rollofft((cbuf[ix] & 0x18) >> 3);
				fe_delivery_system_t s = dvbcompat::system2delsyst((cbuf[ix] & 0x04) >> 2);
				fe_modulation_t      m = dvbcompat::modulation2modulationt((cbuf[ix] & 0x03));

				dprintf(25, "POLARISATION: %s\n", dvbcompat::voltaget2cstr(v));
				dprintf(25, "ROLL OFF: %s\n",     dvbcompat::rollofft2cstr(r));
				dprintf(25, "DELSYS: %s\n",       dvbcompat::delsyst2cstr(s));
				dprintf(25, "MODULATION: %s\n",   dvbcompat::modulationt2cstr(m));
				ix += 1;

				if (delsys == SYS_DVBS2)
					rolloff = r;

				/* Symbol rate */
				ss.str("");
				ss << setfill(' ') << hex << (int)cbuf[ix++];
				ss << setfill('0');
				ss << hex << setw(2) << (int)cbuf[ix++];
				ss << hex << setw(2) << (int)cbuf[ix++];
				ss << hex << setw(2) << (int)(cbuf[ix++] & 0xf0) << '0';
				dprintf(25, "SRATE: %s\n", ss.str().c_str());
			
				/* Forward error correction */
				fe_code_rate_t cr = dvbcompat::fec_inner2coderatet(cbuf[ix+3] & 0x0f);
				dprintf(25, "FEC: %s\n", dvbcompat::coderatet2cstr(cr));
				}
				break;

			default:
				if (tag_id < 0x80)
					dprintf(20, "UNHANDLED TAG: 0x%02x\n", tag_id);
		}
	}
        else
		dprintf(20, "UNHANDLED TABLE: 0x%02x\n", tableid);
}

void
Transponder::readdesc(unsigned char tableid, unsigned char *cbuf, int info_length)
{
	dprintf(20, "tableid = %d, cbuf = %s, info_length = %d\n",
		tableid, (cbuf == NULL) ? "(null)" : "(...)", info_length);

	int ix = 0;
	while (ix < info_length)
	{
		unsigned char tag_id = cbuf[ix++];
		unsigned char tag_len = cbuf[ix++];
		dprintf(20, "tag_id = 0x%02x, tag_len = %d\n", tag_id, tag_len);
		if (tag_len)
		{
			addtag(tableid, cbuf, ix, tag_id, tag_len);
			ix += tag_len;
		}
	}
}

unsigned int
Transponder::findservice(int program_number)
{
	for (unsigned int serviceix = 0; serviceix < service.size(); serviceix++)
	{
		if (service[serviceix].sid == program_number)
			return serviceix;
	}
	return Service::MAX;
}

unsigned int
Transponder::addservice(int program_number, int pmt_pid)
{
	unsigned int serviceix = findservice(program_number);
	if (serviceix == Service::MAX)
	{
		service.push_back(Service(program_number, pmt_pid));
		serviceix = findservice(program_number);
	}
	else
		service[serviceix].refcount++;
	return serviceix;
}

unsigned int
Transponder::findpmt(int pmt_pid)
{
	for (unsigned int pmtix = 0; pmtix < pmt.size(); pmtix++)
	{
		if (pmt[pmtix].pid == pmt_pid)
			return pmtix;
	}
	return Service::MAX;
}

unsigned int
Transponder::addpmt(int pmt_pid)
{
	unsigned int pmtix = findpmt(pmt_pid);
	if (pmtix == Service::MAX)
	{
		pmt.push_back(Pmt(pmt_pid));
		pmtix = findpmt(pmt_pid);
	}
	else
		pmt[pmtix].refcount++;
	return pmtix;
}

int
Transponder::readsdt(class Satellite *sat)
{
	sat->lnb.start(0x11, SDT);

	unsigned int section_number = 0, last_section_number;
	do
	{
		unsigned char cbuf[io::BUFLEN];

		int section_length = read(sat, cbuf, SDT, section_number, &last_section_number);
		if (section_length <= 0)
		{
			dprintf(20, "SDT(0x%02x) TIMEOUT\n", SDT);
			sat->lnb.stop();
			return -ETIMEDOUT;
		}

		unsigned int transport_stream_id = ID_16(cbuf,3);
		dprintf(20, "TRANSPORT STREAM ID = %d (0x%04x)\n", transport_stream_id, transport_stream_id);

		unsigned int original_network_id = ID_16(cbuf,8);
		dprintf(20, "ORIGINAL NETWORK ID = %d (0x%04x)\n", original_network_id, original_network_id);

		if (transport_stream_id != tid)
			dprintf(20, "TRANSPORT STREAM ID = %d (0x%04x) != PAT TID %d (%04x)\n",
				transport_stream_id, transport_stream_id, tid, tid);

		onid = original_network_id;
		tid = transport_stream_id;
			
		int ix = 11;
		while (ix < section_length)
		{
			unsigned int service_id = ID_16(cbuf, ix);
			unsigned int serviceix = findservice(service_id);

			if (serviceix == Service::MAX)
			{
				dprintf(20, "SERVICE ID = %d (0x%04x) (NOT FOUND)\n",
					service_id, service_id);
				serviceix = addservice(service_id);
			}
			else
				dprintf(20, "SERVICE ID = %d (0x%04x) (FOUND)\n",
					service_id, service_id);

			ix += 3;
			unsigned int info_length = INFO_LENGTH(cbuf, ix);
			ix += 2;
			if (info_length)
			{
				service[serviceix].readdesc(SDT, &cbuf[ix], info_length, service[serviceix].pmt_pid);
				ix += info_length;
			}
		}
	}
	while (section_number++ < last_section_number);
				
	sat->lnb.stop();

	return 0;
}

int
Transponder::readpmt(class Satellite *sat)
{
	int ret = 0;

	for (unsigned int pmtix = 0; pmtix < pmt.size(); pmtix++)
	{
		ret = 0;
		dprintf(20, "PMT(0x%02x) PID = %d REFCOUNT = %d\n", PMT, pmt[pmtix].pid, pmt[pmtix].refcount);

		sat->lnb.start(pmt[pmtix].pid, PMT);
		for (unsigned int pmtref = 0; pmtref < pmt[pmtix].refcount && ret != -ETIMEDOUT; pmtref++)
		{
			unsigned int section_number = 0, last_section_number;
			do
			{
				unsigned char cbuf[io::BUFLEN];

				int section_length = read(sat, cbuf, PMT, section_number, &last_section_number);
				if (section_length <= 0)
				{
					dprintf(20, "PMT(0x%02x) TIMEOUT\n", PMT);
					ret = -ETIMEDOUT;
					break;
				}

				unsigned int program_number = ID_16(cbuf, 3);
				unsigned int serviceix = findservice(program_number);

				if (serviceix == Service::MAX)
				{
					dprintf(20, "PROGRAM NUMBER = %d (0x%04x) (NOT FOUND)\n",
						program_number, program_number);
					serviceix = addservice(program_number, pmt[pmtix].pid);
				}
				else
					dprintf(20, "PROGRAM NUMBER = %d (0x%04x) (FOUND)\n",
						program_number, program_number);
	
				service[serviceix].pcr_pid = ID_13(cbuf, 8);
				dprintf(20, "PCR PID = %d (0x%04x)\n", service[serviceix].pcr_pid, service[serviceix].pcr_pid);
	
				int ix = 10;
				unsigned int info_length = INFO_LENGTH(cbuf, ix);
				ix += 2;
				if (info_length)
				{
					service[serviceix].readdesc(PMT, &cbuf[ix], info_length, service[serviceix].pmt_pid);
					ix += info_length;
				}
				while (ix < section_length)
				{
					unsigned int stream_type = cbuf[ix++];
					unsigned int elementary_pid = ID_13(cbuf, ix);
							
					dprintf(20, "PROGRAM NUMBER = %d, STREAM TYPE = %d, ELEMENTARY PID = %d\n",
						program_number, stream_type, elementary_pid);
	 
					service[serviceix].addstream(stream_type, elementary_pid);

					ix += 2;
					info_length = INFO_LENGTH(cbuf, ix);
					ix += 2;
					if (info_length)
					{
						service[serviceix].readdesc(PMT, &cbuf[ix], info_length, elementary_pid);
						ix += info_length;
					}
				}
			}
			while (section_number++ < last_section_number);
		}
		sat->lnb.stop();
	}
	return ret;
}

/*
 * NIT Actual/Other
 */
int
Transponder::readnit(class Satellite *sat, unsigned char tableid)
{
	sat->lnb.start(nit, tableid);

	unsigned int section_number = 0, last_section_number;
	do
	{
		unsigned char cbuf[io::BUFLEN];

		int section_length = read(sat, cbuf, ANIT, section_number, &last_section_number);
		if (section_length <= 0)
		{
			dprintf(20, "NIT(%02x/%02x) TIMEOUT\n", nit, tableid);
			sat->lnb.stop();
			return -ETIMEDOUT;
		}
		dprintf(20, "NIT SECTION LENGTH %d\n", section_length);
		
		unsigned int network_id = ID_16(cbuf,3);
		dprintf(20, "NETWORK ID = %d (0x%04x)\n", network_id, network_id);
		
		nid = network_id;

		int ix = 8;
		unsigned int info_length = INFO_LENGTH(cbuf, ix);
		ix += 2;
		if (info_length)
		{
			readdesc(tableid, &cbuf[ix], info_length);
			ix += info_length;
		}
		ix += 2;
		while (ix < section_length)
		{
			unsigned int transport_stream_id = ID_16(cbuf, ix + 0);
			dprintf(20, "TRANSPORT STREAM ID = %d (0x%04x)\n", transport_stream_id, transport_stream_id);

			unsigned int original_network_id = ID_16(cbuf, ix + 2);
			dprintf(20, "ORIGINAL NETWORK ID = %d (0x%04x)\n", original_network_id, original_network_id);

			ix += 4;
			info_length = INFO_LENGTH(cbuf, ix);
			ix += 2;
			if (original_network_id == onid && transport_stream_id == tid)
				readdesc(tableid, &cbuf[ix], info_length);
			ix += info_length;
		}
	}
	while (section_number++ < last_section_number);
		
	sat->lnb.stop();

	return 0;
}

int
Transponder::readpat(class Satellite *sat)
{
	sat->lnb.start(0x00, PAT);

	unsigned int section_number = 0, last_section_number;
	do
	{
		unsigned char cbuf[io::BUFLEN];

		int section_length = read(sat, cbuf, PAT, section_number, &last_section_number);
		if (section_length <= 0)
		{
			dprintf(20, "PAT(%02x) TIMEOUT\n", PAT);
			sat->lnb.stop();
			return -ETIMEDOUT;
		}

		unsigned int transport_stream_id = tid = ID_16(cbuf, 3);
		dprintf(20, "TRANSPORT STREAM ID = %d (0x%04x)\n", transport_stream_id, transport_stream_id);

		for (int ix = 8; ix < section_length; ix += 4)
		{
			unsigned int program_number = ID_16(cbuf, ix);
			if (program_number == 0)
			{
				unsigned int network_pid = ID_13(cbuf, ix + 2);
				dprintf(20, "PROGRAM NUMBER = %d (0x%04x), NETWORK PID = %d (0x%04x)\n",
					program_number, program_number, network_pid, network_pid);

				if (network_pid > 0 && network_pid < 65535)
					nit = network_pid;
			}
			else
			{
				unsigned int pmt_pid = ID_13(cbuf, ix + 2);
				dprintf(20, "PROGRAM NUMBER = %d (0x%04x), PMT PID = %d (0x%04x)\n",
					program_number, program_number, pmt_pid, pmt_pid);

				addpmt(pmt_pid);
				addservice(program_number, pmt_pid);
			}
		}
	}
	while (section_number++ < last_section_number);
		
	sat->lnb.stop();

	return service.size();
}
