/*
 * dvb2010 for linux-dvb
 *
 * Copyright (c) 2008 Darron Broad
 * All rights reserved.
 *
 * Licensed under the terms of the BSD license, see file LICENSE for details.
 */

#include "dvb2010.h"
#include "cstring.h"
#include "util.h"
#include "lnb.h"
#include "satellite.h"
#include "transponder.h"
#include "service.h"
#include "xmlscan.h"

void
usage(char *execname, string errstr, int errnum)
{
	cerr << "USAGE: " << execname << " [OPTIONS] POSITION [FREQUENCY]\n\n";
	cerr << "Interface options:\n"
		" -a num\t\tDVB-S/S2 adapter num.\n"
		" -s num\t\tDiSEqC switch port num (range: 0..3).\n"
		" -d n,b\t\tDiSEqC repeat count n, burst disable (0) or\n\t\tburst enable (1).\n"
		" -l l,h,s\tLNBF lo-band LO l, hi-band LO h & switch-over\n\t\tfreq s in MHz.\n"
		" -p pol\t\tLNBF polarisation, linear (l) or circular (c).\n"
		" -w sec\t\tWait up to sec seconds for tuning lock.\n\n"
		" (default)\t-a 0 -s 0 -d 2,1 -l 9750,10600,11700 -p l -w 5\n\n"
		"Tuning args:\n"
		" POSITION\tSatellite position (range: 179.9W..180.0E).\n"
		" FREQUENCY\tTransponder frequency in MHz.\n\n";
	cerr << "FILES:\n"
		" " << Satellite::SATXML << " \n\t\tSatellite position and transponder parameters\n"
		"\t\t(Eg. http://satellites-xml.eu/).\n\n" <<
		" " << getenv("HOME") << "/.dvb2010/config\n\t\tSQLite2/3 or MySQL5 database configuration.\n\n";

	cerr << "EXAMPLES:\n"
		" " << execname << " -a 1 -s 1 19.2E\n\t\tScan 19.2E using DVB adapter 1, DiSEqC port 1.\n\n";
	cerr << " " << execname << " 28.2E 10714\n\t\tScan 10714MHz at 28.2E.\n";

	if (errstr != "")
		cerr << "\nERROR:\n " << errstr << ".\n";

	exit(errnum);
}

int
main(int argc, char **argv)
{
	char c, *execdup, *execname, *comma;
	Satellite sat;

	execdup = strdup(argv[0]);
	execname = basename(execdup);

	opterr = 0;
	while ((c = getopt(argc, argv, "a:s:d:l:p:w:")) != -1)
	{
		switch (c)
		{
			case 'a':
				sat.lnb.phy_adapter = util::toInt(optarg);
				break;
			case 's':
				sat.lnb.dsec_port = util::toInt(optarg);
				break;
			case 'd':
				sat.lnb.dsec_repeat = util::toInt(optarg);
				comma = strstr(optarg, ",");
				if (comma != NULL)
					sat.lnb.dsec_burst = util::toInt(++comma);
				break;
			case 'l':
				sat.lnb.low = 1000 * util::toInt(optarg);
				sat.lnb.high = sat.lnb.threshold = 0;
				comma = strstr(optarg, ",");
				if (comma != NULL)
				{
					sat.lnb.high = 1000 * util::toInt(++comma);
					comma = strstr(comma, ",");
					if (comma != NULL)
						sat.lnb.threshold = 1000 * util::toInt(++comma);
				}
				break;
			case 'p':
				sat.lnb.polarisation = toupper(optarg[0]);
				if (sat.lnb.polarisation != 'L' && sat.lnb.polarisation != 'C')
				{
					string error = "Invalid polarisation " + (string)optarg;
					usage(execname, error, EX_DATAERR);
				}
				break;
			case 'w':
				sat.lnb.tuning_delay = util::toInt(optarg);
				break;
			case '?':
				usage(execname, "", EX_OK);
				break;
			default:
				string error = "Unknown option " + (string)optarg;
				usage(execname, error, EX_DATAERR);
				break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1)
		usage(execname, "Missing args(s)", EX_DATAERR);
	if (argc > 2)
		usage(execname, "Too many args(s)", EX_DATAERR);

	sat.lnb.position = util::cstringToAngle(argv[0]);
	if (sat.lnb.position == INT_MAX)
	{
		string error = "Invalid satellite position " + (string)argv[0];
		usage(execname, error, EX_DATAERR);
	}

	if (argc >= 2)
		sat.lnb.frequency = util::toInt(argv[1]);

	setlocale(LC_ALL, "");
	try
	{
		sat.lnb.open(0);
		sat.scan();
	}
	catch (const char *e1)
	{
		cerr << "\nError: " << e1 << endl;
	}
	catch (string e2)
	{
		cerr << "\nError: " << e2 << endl;
	}
	sat.lnb.close();

	exit(EX_OK);
}
